import cartopy.crs as ccrs
import cartopy.io.img_tiles as cimgt
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
from typing import Tuple


def plot_locations_osm(
    gps_series: gpd.GeoSeries,
    c_series: np.ndarray = None,
    zoom_level: int = 10,
    scatter_size: int = 10,
    fig_size: Tuple[int, int] = (12, 10),
    save_path=None,
    show=True,
    **kwargs
):
    """
    Plot a series of gps locations onto a OSM map

    :param gps_series: geopandas series of the gps value in EPSG:4326 to plot
    :param c_series: series of values to be interpreted as colors in the scatterplot
    :param scatter_size: scalar or series to be interpreted as size in the scatterplot
    :param zoom_level: zoom level for resolution of OSM tiles
    :param fig_size: the figure size of the generated plot
    :param **kwargs: additional keword arguements which will directly be passed on to scatterplot
    :param save_path: if not none then the plot will be saved under this path
    """

    if gps_series.crs != "EPSG:4326":
        raise ValueError("Make sure to pass data with EPSG:4326 projection")

    request = cimgt.OSM()
    extent = gps_series.total_bounds[[0, 2, 1, 3]]
    plt.figure(figsize=fig_size)
    ax = plt.axes(projection=request.crs)
    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")
    ax.set_extent(extent)
    ax.add_image(request, zoom_level)
    if c_series is not None:
        ax.scatter(
            gps_series.x,
            gps_series.y,
            c=c_series,
            transform=ccrs.PlateCarree(),
            s=scatter_size,
            **kwargs
        )
    else:
        ax.scatter(
            gps_series.x,
            gps_series.y,
            transform=ccrs.PlateCarree(),
            s=scatter_size,
            **kwargs
        )
    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight")
    if show:
        plt.show()


def plot_street_nodes(
    street_series: gpd.GeoSeries, axs, marker: str = "o", color: str = "blue"
):
    """
    Plots the street_series with nodes along each way.

    :param street_series: geopandas series of LineStrings
    :param axs: denotes the axis on which to plot
    :param marker: marker for the nodes
    :param color: color of the plotted lines and nodes
    """
    for linestring in street_series.geometry:
        axs.plot(
            np.array(linestring.xy[0]),
            np.array(linestring.xy[1]),
            marker=marker,
            color=color,
        )
