import asyncio
from datetime import datetime
from typing import List, Optional, Tuple

import aiohttp
import geopandas as gpd
import pandas as pd
import requests
from multidict import MultiDict
from tqdm import tqdm
from tqdm.asyncio import tqdm as tqdm_async

BASE_URL = "https://www.netztest.at/opendata"
SUBDOMAIN_DETAILS = "opentests"
SUBDOMAIN_OVERVIEW = "opentests/search"


def rtr_details(open_test_uuids: List[str]) -> List[dict]:
    """
    Fetch test details from RTR-Opendata for a list of open_test_uuids.
    These open_test_uuids can for instance be obtained via data_extractor.fetch_rtr_overview()
    Detailed information about the API endpoint and the meaning of returned parameters is available under:
    https://www.netztest.at/en/OpenDataSpecification.html.

    :param open_test_uuids: List of open_test_uuids for which test details will be fetched

    :return: A list of dictionaries with raw test results
    """

    async def fetch(session, url):
        async with session.get(url) as response:
            return await response.json()

    async def query(urls):
        tasks = []
        async with aiohttp.ClientSession() as session:
            # Generate a task for each URL to fetch
            tasks = [asyncio.create_task(fetch(session, url)) for url in urls]

            # Run tasks on the event loop and print progress via TQDM
            return [await f for f in tqdm_async.as_completed(tasks)]

    urls = [f"{BASE_URL}/{SUBDOMAIN_DETAILS}/{uuid}" for uuid in open_test_uuids]



    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:  # no event loop running:
        loop = asyncio.get_event_loop()
    
    results = loop.run_until_complete(query(urls))

    return results


def rtr_overview(
    time_min: Optional[datetime] = None,
    time_max: Optional[datetime] = None,
    gps_boundaries: Optional[gpd.GeoSeries] = None,
    cat_technology: str = "4G",
    max_results: int = 1000,
    raw_params: List[Tuple[str, str]] = [],
    verbose: bool = False
) -> pd.DataFrame:
    """
    Queries the RTR-Netztest API according to the passed filters.
    Detailed information about the API endpoint and the meaning of returned parameters is available under:
    https://www.netztest.at/en/OpenDataSpecification.html.
    Comprehensive results for each test can be obtained by passing the
    open_test_uuids of the fetched measurements to fetch_rtr_details.

    :param time_min: Lower bound for time-filter. Passed Datetime objects will be interpreted as UTC.
    :param time_max: Upper bound for time-filter. Passed Datetime objects will be interpreted as UTC.

    :gps_boundaries: GeoSeries in EPSG:4326 of the boundaries to filter for.
    :cat_technology: CatTechnology to filter for. Examples are 3G and 4G.

    :max_results: Upper limit of measurements to fetch.

    :raw_params: Passed raw parameter tuples will be appended to the request url as key, value pairs.

    :return: A pandas dataframe consisting of all the fetched measurements.
    """

    url = f"{BASE_URL}/{SUBDOMAIN_OVERVIEW}"
    request_params = MultiDict()

    # Build the request parameters

    if gps_boundaries is not None:
        if gps_boundaries.crs != "EPSG:4326":
            raise ValueError("Make sure to pass data with EPSG:4326 projection")
        else:
            long_min, lat_min, long_max, lat_max = tuple(gps_boundaries.total_bounds)
            request_params.add("long", f">{long_min}")
            request_params.add("long", f"<{long_max}")
            request_params.add("lat", f">{lat_min}")
            request_params.add("lat", f"<{lat_max}")

    if time_min is not None:
        time_min = round(time_min.timestamp() * 1000)
        request_params.add("time", f">{time_min}")

    if time_max is not None:
        time_max = round(time_max.timestamp() * 1000)
        request_params.add("time", f"<{time_max}")

    request_params.add("cat_technology", cat_technology)

    if max_results > 500:
        request_params.add("max_results", 500)
    else:
        request_params.add("max_results", max_results)

    for param, value in raw_params:
        request_params.add(param, value)

    request_params_list = list(request_params.items())

    results = []
    test_counter = 0
    pbar = tqdm(total=max_results, desc="ETA Assuming max_results available:")
    with requests.Session() as session:
        # Initial Request
        response = session.get(url, params=request_params_list)
        if verbose:
            print(f"Requesting URL: {response.request.url}")
        response_json = response.json()

        tests = response_json["results"]
        results += tests
        pbar.update(len(tests))
        test_counter += len(tests)

        # Repeated Requests to deal with paging mechanism of API
        while (next_cursor := response_json.get("next_cursor")) is not None:

            request_params["cursor"] = next_cursor
            request_params_list = list(request_params.items())

            response = session.get(url, params=request_params_list)
            response_json = response.json()

            tests = response_json["results"]
            results += tests
            pbar.update(len(tests))

            test_counter += len(tests)

            if test_counter >= max_results:
                break

    pbar.close()

    return pd.DataFrame(results)
